@@ -36,7 +36,7 @@
 
 #define TMP_SIZE 512
 
-/* #define DEBUG_BRANCH 1 */
+//#define DEBUG_BRANCH 1
 
 static int slot_bits[4] = {
     FS_BIND_MODEL,
@@ -170,7 +170,9 @@ fs_rid_vector **fs_bind(fs_backend *be, fs_segment segment, unsigned int tobind,
     double then = fs_time();
 
     limit = (limit == -1) ? INT_MAX : limit;
-    offset = (offset == -1) ? 0 : offset;
+    if (offset != -1 && offset != 0) {
+	fs_error(LOG_ERR, "fs_bind called with non-zero offset");
+    }
 
     int cols = 0;
     for (int i=0; i<4; i++) {
@@ -366,17 +368,16 @@ fs_error(LOG_INFO, "bind() branch");
 	    }
 	}
     /* query like (_ s p _) */
-    } else if (fs_rid_vector_length(pv) != 0) {
-	if (tobind & FS_BIND_BY_SUBJECT) {
+    } else if (tobind & FS_BIND_BY_SUBJECT && fs_rid_vector_length(pv) > 0 && fs_rid_vector_length(sv) > 0) {
+	if (fs_rid_vector_length(pv) == 1) {
 #ifdef DEBUG_BRANCH
 fs_error(LOG_INFO, "bind() branch");
 #endif
 	    const int ml = mvl ? mvl : 1;
 	    const int ol = ovl ? ovl : 1;
-	    for (int p=0; p<pvl && count<limit; p++) {
-		fs_ptree *pt = fs_backend_get_ptree(be, pv->data[p], 0);
-		if (!pt) continue;
-		for (int s=0; s<svl; s++) {
+	    fs_ptree *pt = fs_backend_get_ptree(be, pv->data[0], 0);
+	    if (pt) {
+		for (int s=0; s<svl && count<limit; s++) {
 		    fs_rid pk = sv->data[s];
 		    for (int m=0; m<ml && count<limit; m++) {
 			for (int o=0; o<ol; o++) {
@@ -386,7 +387,7 @@ fs_error(LOG_INFO, "bind() branch");
 			    fs_ptree_it *it = fs_ptree_search(pt, pk, pair);
 			    while (it && fs_ptree_it_next(it, pair) && count<limit) {
 				const fs_rid quad[4] =
-				    { pair[0], pk, pv->data[p], pair[1] };
+				    { pair[0], pk, pv->data[0], pair[1] };
 				if (!bind_same(quad, tobind)) continue;
 				if (!graph_ok(quad, tobind)) continue;
 				count++;
@@ -402,13 +403,45 @@ fs_error(LOG_INFO, "bind() branch");
 fs_error(LOG_INFO, "bind() branch");
 #endif
 	    const int ml = mvl ? mvl : 1;
-	    const int sl = svl ? svl : 1;
+	    const int ol = ovl ? ovl : 1;
 	    for (int p=0; p<pvl && count<limit; p++) {
-		fs_ptree *pt = fs_backend_get_ptree(be, pv->data[p], 1);
+		fs_ptree *pt = fs_backend_get_ptree(be, pv->data[p], 0);
 		if (!pt) continue;
+		/* we have to use values from s and p for same row */
+		fs_rid pk = sv->data[p];
+		for (int m=0; m<ml && count<limit; m++) {
+		    for (int o=0; o<ol; o++) {
+			fs_rid pair[2] = { FS_RID_NULL, FS_RID_NULL };
+			if (mvl) pair[0] = mv->data[m];
+			if (ovl) pair[1] = ov->data[o];
+			fs_ptree_it *it = fs_ptree_search(pt, pk, pair);
+			while (it && fs_ptree_it_next(it, pair) && count<limit) {
+			    const fs_rid quad[4] =
+				{ pair[0], pk, pv->data[p], pair[1] };
+			    if (!bind_same(quad, tobind)) continue;
+			    if (!graph_ok(quad, tobind)) continue;
+			    count++;
+			    bind_results(quad, tobind, ret);
+			}
+			fs_ptree_it_free(it);
+		    }
+		}
+	    }
+	}
+    /* query like (_ _ p o) */
+    } else if (tobind & FS_BIND_BY_OBJECT && fs_rid_vector_length(pv) > 0 && fs_rid_vector_length(ov) > 0) {
+#ifdef DEBUG_BRANCH
+fs_error(LOG_INFO, "bind() branch");
+#endif
+	/* optimisation for exactly one predicate */
+	if (pvl == 1) {
+	    const int ml = mvl ? mvl : 1;
+	    const int sl = svl ? svl : 1;
+	    fs_ptree *pt = fs_backend_get_ptree(be, pv->data[0], 1);
+	    if (pt) {
 		for (int o=0; o<ovl && count<limit; o++) {
 		    fs_rid pk = ov->data[o];
-		    for (int m=0; m<ml; m++) {
+		    for (int m=0; m<ml && count<limit; m++) {
 			for (int s=0; s<sl && count<limit; s++) {
 			    fs_rid pair[2] = { FS_RID_NULL, FS_RID_NULL };
 			    if (mvl) pair[0] = mv->data[m];
@@ -416,7 +449,7 @@ fs_error(LOG_INFO, "bind() branch");
 			    fs_ptree_it *it = fs_ptree_search(pt, pk, pair);
 			    while (it && fs_ptree_it_next(it, pair) && count<limit) {
 				const fs_rid quad[4] = {
-				    pair[0], pair[1], pv->data[p], pk
+				    pair[0], pair[1], pv->data[0], pk
 				};
 				if (!bind_same(quad, tobind)) continue;
 				if (!graph_ok(quad, tobind)) continue;
@@ -428,6 +461,33 @@ fs_error(LOG_INFO, "bind() branch");
 		    }
 		}
 	    }
+	} else {
+	    const int ml = mvl ? mvl : 1;
+	    const int sl = svl ? svl : 1;
+	    for (int p=0; p<pvl && count<limit; p++) {
+		fs_ptree *pt = fs_backend_get_ptree(be, pv->data[p], 1);
+		if (!pt) continue;
+		/* we need values from teh same row for o and p */
+		fs_rid pk = ov->data[p];
+		for (int m=0; m<ml; m++) {
+		    for (int s=0; s<sl && count<limit; s++) {
+			fs_rid pair[2] = { FS_RID_NULL, FS_RID_NULL };
+			if (mvl) pair[0] = mv->data[m];
+			if (svl) pair[1] = sv->data[s];
+			fs_ptree_it *it = fs_ptree_search(pt, pk, pair);
+			while (it && fs_ptree_it_next(it, pair) && count<limit) {
+			    const fs_rid quad[4] = {
+				pair[0], pair[1], pv->data[p], pk
+			    };
+			    if (!bind_same(quad, tobind)) continue;
+			    if (!graph_ok(quad, tobind)) continue;
+			    count++;
+			    bind_results(quad, tobind, ret);
+			}
+			fs_ptree_it_free(it);
+		    }
+		}
+	    }
 	}
     } else {
 	if (tobind & FS_BIND_BY_SUBJECT) {
@@ -670,55 +730,10 @@ fs_rid_vector **fs_bind_first(fs_backend *be, fs_segment segment,
 			      fs_rid_vector *pv, fs_rid_vector *ov,
                               int count)
 {
-    if (!(tobind & (FS_BIND_BY_SUBJECT | FS_BIND_BY_OBJECT))) {
-	fs_error(LOG_ERR, "tried to bind without s/o spec");
-
-	return NULL;
-    } else if ((tobind & (FS_BIND_BY_SUBJECT | FS_BIND_BY_OBJECT)) ==
-	       (FS_BIND_BY_SUBJECT | FS_BIND_BY_OBJECT)) {
-	fs_error(LOG_ERR, "tried to bind with s+o spec set");
-
-	return NULL;
-    }
-    double then = fs_time();
-
-    int cols = 0;
-    for (int i=0; i<4; i++) {
-	if (tobind & slot_bits[i]) {
-	    cols++;
-	}
-    }
-
-    /* streaming binds aren't implemented in this branch yet */
+    /* streaming binds aren't implemented in this branch */
     fs_error(LOG_ERR, "not implemented");
+
     return NULL;
-
-    fs_rid_vector **ret;
-    if (cols == 0) {
-	ret = calloc(1, sizeof(fs_rid_vector *));
-    } else {
-	ret = calloc(cols, sizeof(fs_rid_vector *));
-    }
-    for (int i=0; i<cols; i++) {
-	ret[i] = fs_rid_vector_new(0);
-    }
-
-    if (be->stream) {
-	fs_error(LOG_ERR, "bind_first(%d) while already streaming", segment);
-	return NULL;
-    }
-
-    /* TODO bind streaming quad pattern */
-
-    be->out_time[segment].bind_count++;
-    be->out_time[segment].bind += fs_time() - then;
-
-    if (cols == 0) {
-	free(ret);
-	return NULL;
-    }
-
-    return ret;
 }
 
 fs_rid_vector **fs_bind_next(fs_backend *be, fs_segment segment,
