@@ -21,6 +21,8 @@
 #include <math.h>
 #include <string.h>
 #include <pcre.h>
+#include <time.h>
+#include <errno.h>
 
 #include "filter-datatypes.h"
 #include "filter.h"
@@ -28,6 +30,7 @@
 #include "../common/4s-hash.h"
 #include "../common/error.h"
 #include "../common/rdf-constants.h"
+#include "../libs/mt19937-64/mt64.h"
 
 typedef struct _fs_date_fields {
     int year;
@@ -122,6 +125,9 @@ static fs_value cast_double(fs_value a)
     if (a.lex && strlen(a.lex)) {
         char *end = NULL;
         a.fp = strtod(a.lex, &end);
+        /* this is bad, but some code somewhere is checking errno when a
+         * function is returning sucess and getting tripped up by this */
+        errno = 0;
         if (*end == '\0') {
             a.valid |= fs_valid_bit(FS_V_FP);
 
@@ -248,7 +254,7 @@ fs_value fn_plus(fs_query *q, fs_value a)
 fs_value_print(a);
 #endif
     if (fs_is_numeric(&a)) {
-	return a;
+        return a;
     }
 
     return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:plus");
@@ -257,22 +263,226 @@ fs_value_print(a);
 fs_value fn_minus(fs_query *q, fs_value a)
 {
     if (fs_is_numeric(&a)) {
-	if (a.attr == fs_c.xsd_double || a.attr == fs_c.xsd_float) {
-	    a.fp = -a.fp;
-	} else if (a.attr == fs_c.xsd_decimal) {
+        if (a.attr == fs_c.xsd_double || a.attr == fs_c.xsd_float) {
+            a.fp = -a.fp;
+        } else if (a.attr == fs_c.xsd_decimal) {
             fs_decimal_negate(&a.de, &a.de);
-	} else if (a.attr == fs_c.xsd_integer) {
-	    a.in = -a.in;
-	} else {
-	    return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:minus");
-	}
+        } else if (a.attr == fs_c.xsd_integer) {
+            a.in = -a.in;
+        } else {
+            return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:minus");
+	    }
 
-	return a;
+        if (a.lex != NULL) {
+            a.lex = NULL;
+        }
+        return a;
     }
 
     return fs_value_error(FS_ERROR_INVALID_TYPE, "non-numeric arguments to fn:minus");
 }
 
+fs_value fn_numeric_abs(fs_query *q, fs_value a)
+{
+    if (!fs_is_numeric(&a)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "non-numeric argument to fn:abs");
+    }
+
+    if (a.attr == fs_c.xsd_double || a.attr == fs_c.xsd_float) {
+        a.fp = fabs(a.fp);
+    } else if (a.attr == fs_c.xsd_decimal) {
+        if (fs_decimal_less_than(&a.de, fs_decimal_zero)) {
+            fs_decimal_negate(&a.de, &a.de);
+        }
+    } else if (a.attr == fs_c.xsd_integer) {
+        if (a.in < 0) {
+            a.in = -a.in;
+        }
+    } else {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:abs");
+    }
+
+    if (a.lex != NULL) {
+        a.lex = NULL;
+    }
+
+    return a;
+}
+
+fs_value fn_numeric_floor(fs_query *q, fs_value a)
+{
+    if (!fs_is_numeric(&a)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "non-numeric argument to fn:floor");
+    }
+
+    if (a.attr == fs_c.xsd_integer) {
+        /* do nothing for integers */
+        return a;
+    }
+
+    if (a.attr == fs_c.xsd_double || a.attr == fs_c.xsd_float) {
+        a.fp = floor(a.fp);
+    } else if (a.attr == fs_c.xsd_decimal) {
+        fs_decimal r;
+
+        if (fs_decimal_less_than(&a.de, fs_decimal_zero)) {
+            fs_decimal one;
+            fs_decimal_init_from_double(&one, (double)1.0);
+            fs_decimal_subtract(&a.de, &one, &r);
+        }
+        else {
+            fs_decimal_copy(&a.de, &r);
+        }
+
+        for (int i=FS_D_OVER_DIGITS+FS_D_INT_DIGITS; i < FS_D_DIGITS; i++) {
+            r.digit[i] = 0;
+        }
+
+        fs_decimal_copy(&r, &a.de);
+    } else {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:floor");
+    }
+
+    if (a.lex != NULL) {
+        a.lex = NULL;
+    }
+
+    return a;
+}
+
+fs_value fn_numeric_ceil(fs_query *q, fs_value a)
+{
+    if (!fs_is_numeric(&a)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "non-numeric argument to fn:ceil");
+    }
+
+    if (a.attr == fs_c.xsd_integer) {
+        /* do nothing for integers */
+        return a;
+    }
+
+    if (a.attr == fs_c.xsd_double || a.attr == fs_c.xsd_float) {
+        a.fp = ceil(a.fp);
+    } else if (a.attr == fs_c.xsd_decimal) {
+        fs_decimal r;
+
+        if (fs_decimal_greater_than(&a.de, fs_decimal_zero)) {
+            fs_decimal one;
+            fs_decimal_init_from_double(&one, (double)1.0);
+            fs_decimal_add(&a.de, &one, &r);
+        }
+        else {
+            fs_decimal_copy(&a.de, &r);
+        }
+
+        for (int i=FS_D_OVER_DIGITS+FS_D_INT_DIGITS; i < FS_D_DIGITS; i++) {
+            r.digit[i] = 0;
+        }
+
+        fs_decimal_copy(&r, &a.de);
+    } else {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:ceil");
+    }
+
+    if (a.lex != NULL) {
+        a.lex = NULL;
+    }
+
+    return a;
+}
+
+fs_value fn_numeric_round(fs_query *q, fs_value a)
+{
+    if (!fs_is_numeric(&a)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "non-numeric argument to fn:round");
+    }
+
+    if (a.attr == fs_c.xsd_integer) {
+        /* do nothing for integers */
+        return a;
+    }
+
+    if (a.attr == fs_c.xsd_double || a.attr == fs_c.xsd_float) {
+        if (a.fp > 0.0) {
+            /* rounding +ve numbers same as c round */
+            a.fp = round(a.fp);
+        }
+        else if (a.fp < 0.0) {
+            /* -2.5 should round to -2.0 */
+            double i;
+            double f = modf(a.fp, &i);
+            a.fp -= f;
+            if (f < -0.5) {
+                a.fp -= 1.0;
+            }
+        }
+    } else if (a.attr == fs_c.xsd_decimal) {
+        /* 0 = round to num,        e.g. 2.x -> 2.0, -2.x -> -2.0
+           1 = round away from num, e.g. 2.x -> 3.0, -2.x -> -3.0 */
+        int round_dir = -1; 
+        int positive = fs_decimal_greater_than_equal(&a.de, fs_decimal_zero);
+
+        int start_pos = FS_D_OVER_DIGITS+FS_D_INT_DIGITS;
+        if ((&a.de)->digit[start_pos] > 5) {
+            round_dir = 1;
+        } else if ((&a.de)->digit[start_pos] < 5) {
+            round_dir = 0;
+        } else {
+            /* first digit should be a 5 */
+            for (int i=start_pos+1; i < FS_D_DIGITS; i++) {
+                if ((&a.de)->digit[i] > 0) {
+                    /* abs(fractional part) is > 0.5 */
+                    round_dir = 1;
+                    break;
+                }
+            }
+
+            if (round_dir == -1) {
+                /* abs(fractional part) == 0.5, round depending on sign */
+                if (positive) {
+                    round_dir = 1;
+                }
+                else {
+                    round_dir = 0;
+                }
+            }
+        }
+
+        fs_decimal one;
+        fs_decimal_init_from_double(&one, (double)1.0);
+        fs_decimal r;
+
+        if (round_dir == 1) {
+            /* Round integral part of number away from zero */
+            if (positive) {
+                fs_decimal_add(&a.de, &one, &r);
+            }
+            else {
+                fs_decimal_subtract(&a.de, &one, &r);
+            }
+        }
+        else {
+            /* Keep integral part of number unchanged */
+            fs_decimal_copy(&a.de, &r);
+        }
+
+        /* Zero all fractional digits of number */
+        for (int i=start_pos; i < FS_D_DIGITS; i++) {
+            r.digit[i] = 0;
+        }
+
+        fs_decimal_copy(&r, &a.de);
+    } else {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, "bad arguments to fn:round");
+    }
+
+    if (a.lex != NULL) {
+        a.lex = NULL;
+    }
+
+    return a;
+}
+
 fs_value fn_numeric_add(fs_query *q, fs_value a, fs_value b)
 {
 #if 0
@@ -1224,6 +1434,7 @@ fs_value fn_substring(fs_query *q, fs_value str, fs_value start, fs_value length
         return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
     }
     start = cast_integer(start);
+    if (start.in < 1) start.in = 1;
     /* 2 arg form */
     if (length.rid == FS_RID_NULL) {
         length = fs_value_integer(INT_MAX);
@@ -1452,4 +1663,64 @@ fs_value fn_contains(fs_query *q, fs_value arg1, fs_value arg2)
     return fs_value_boolean(strstr(arg1.lex, arg2.lex) != NULL);
 }
 
+fs_value fn_rand(fs_query *q)
+{
+    static int seeded = 0;
+    if (!seeded) {
+        init_genrand64(time(NULL));
+        seeded = 1;
+    }
+    
+    fs_value v = fs_value_double(genrand64_real2());
+    v.lex = g_strdup_printf("%.17f", v.fp);
+    fs_query_add_freeable(q, v.lex);
+
+    return v;
+}
+
+#if GLIB_MAJOR_VERSION >= 2 && GLIB_MINOR_VERSION >= 16
+fs_value fn_hash_intl(fs_query *q, fs_value arg, GChecksumType type)
+{
+    if (!fs_is_plain_or_string(arg)) {
+        return fs_value_error(FS_ERROR_INVALID_TYPE, NULL);
+    }
+    arg = fs_value_fill_lexical(q, arg);
+
+    GChecksum *sum = g_checksum_new(type);
+    g_checksum_update(sum, (guchar *)arg.lex, -1);
+    char *str = g_strdup(g_checksum_get_string(sum));
+    g_checksum_free(sum);
+    fs_query_add_freeable(q, str);
+
+    return fs_value_plain(str);
+}
+#endif
+
+fs_value fn_md5(fs_query *q, fs_value arg)
+{
+#if GLIB_MAJOR_VERSION >= 2 && GLIB_MINOR_VERSION >= 16
+    return fn_hash_intl(q, arg, G_CHECKSUM_MD5);
+#else
+    return fs_value_error(FS_ERROR_INVALID_TYPE, "glib version does not support hash functions, at least 2.16.0 required");
+#endif
+}
+
+fs_value fn_sha1(fs_query *q, fs_value arg)
+{
+#if GLIB_MAJOR_VERSION >= 2 && GLIB_MINOR_VERSION >= 16
+    return fn_hash_intl(q, arg, G_CHECKSUM_SHA1);
+#else
+    return fs_value_error(FS_ERROR_INVALID_TYPE, "glib version does not support hash functions, at least 2.16.0 required");
+#endif
+}
+
+fs_value fn_sha256(fs_query *q, fs_value arg)
+{
+#if GLIB_MAJOR_VERSION >= 2 && GLIB_MINOR_VERSION >= 16
+    return fn_hash_intl(q, arg, G_CHECKSUM_SHA256);
+#else
+    return fs_value_error(FS_ERROR_INVALID_TYPE, "glib version does not support hash functions, at least 2.16.0 required");
+#endif
+}
+
 /* vi:set expandtab sts=4 sw=4: */
