@@ -23,6 +23,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <glib.h>
+#include <pthread.h>
 #include <rasqal.h>
 
 #include "4store-config.h"
@@ -47,6 +48,8 @@
 
 #define DEBUG_SIZE(n, thing) printf("@@ %d * sizeof(%s) = %zd\n", n, #thing, n * sizeof(thing))
 
+GStaticMutex rasqal_mutex = G_STATIC_MUTEX_INIT;
+
 static void graph_pattern_walk(fsp_link *link, rasqal_graph_pattern *p, fs_query *q, rasqal_literal *model, int optional, int uni);
 static int fs_handle_query_triple(fs_query *q, int block, rasqal_triple *t);
 static int fs_handle_query_triple_multi(fs_query *q, int block, int count, rasqal_triple *t[]);
@@ -231,6 +234,7 @@ fs_query_state *fs_query_init(fsp_link *link, rasqal_world *rasworld, raptor_wor
         }
     }
 
+    g_static_mutex_lock(&rasqal_mutex);
     if (rasworld) {
         qs->rasqal_world = rasworld;
     } else {
@@ -239,11 +243,18 @@ fs_query_state *fs_query_init(fsp_link *link, rasqal_world *rasworld, raptor_wor
     if (!qs->rasqal_world) {
         fs_error(LOG_ERR, "failed to allocate rasqal world");
     }
+#if RASQAL_VERSION >= 926
+    /* Lower warning level to get only more serious warnings */
+    rasqal_world_set_warning_level(qs->rasqal_world, 25);
+#endif
     if (rasqal_world_open(qs->rasqal_world)) {
         fs_error(LOG_ERR, "failed to intialise rasqal world");
 	fs_query_fini(qs);
+        g_static_mutex_unlock(&rasqal_mutex);
+
 	return NULL;
     }
+    g_static_mutex_unlock(&rasqal_mutex);
     if (rapworld) {
         qs->raptor_world = rapworld;
     } else {
@@ -298,7 +309,6 @@ static void tree_compact(fs_query *q)
 #ifdef DEBUG_MERGE
                 printf("Merge B%d to B%d\n", block, parent);
 #endif
-                done_something = 1;
                 fs_p_vector_append_vector(q->blocks+parent, q->blocks+block);
                 if (q->constraints[parent] == NULL) {
                     q->constraints[parent] = q->constraints[block];
@@ -331,6 +341,7 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
 
     fsp_hit_limits_reset(link);
 
+    g_static_mutex_lock(&rasqal_mutex);
     rasqal_query *rq = rasqal_new_query(qs->rasqal_world, "sparql11", NULL);
     if (!rq) {
         rq = rasqal_new_query(qs->rasqal_world, "laqrs", NULL);
@@ -338,6 +349,7 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
     if (!rq) {
         rq = rasqal_new_query(qs->rasqal_world, "sparql", NULL);
     }
+    g_static_mutex_unlock(&rasqal_mutex);
     if (!rq) {
         fs_error(LOG_ERR, "failed to initialise query system");
 
@@ -358,7 +370,9 @@ fs_query *fs_query_execute(fs_query_state *qs, fsp_link *link, raptor_uri *bu, c
     }
     q->boolean = 1;
     rasqal_world_set_log_handler(q->qs->rasqal_world, q, log_handler);
+    g_static_mutex_lock(&rasqal_mutex);
     int ret = rasqal_query_prepare(rq, (unsigned char *)query, bu);
+    g_static_mutex_unlock(&rasqal_mutex);
     if (ret) {
 	return q;
     }
@@ -835,38 +849,16 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
 #endif
     }
 
-    /* perform all the UNIONs */
-    int first_in_union;
-    for (int i=q->unions; i>0; i--) {
-        first_in_union = 0;
-        for (int j=1; j<=q->block; j++) {
+    /* pick a primary block to hold the result of each UNION operation */
+    /* N.B. union_group 0 indicates no union */
+    int pri_for_union[q->unions+1];
+    pri_for_union[0] = 0;
+    for (int i = 1; i <= q->unions; i++) {
+        pri_for_union[i] = 0;
+        for (int j = q->block; j > 0; j--) {
             if (q->union_group[j] == i) {
-                if (first_in_union == 0) {
-                    first_in_union = j;
-                    if (q->constraints[j]) {
-                        fs_binding *old = q->bb[j];
-                        q->bb[j] = fs_binding_apply_filters(q, j, q->bb[j], q->constraints[j]);
-                        fs_binding_free(old);
-                        raptor_free_sequence(q->constraints[j]);
-                        q->constraints[j] = NULL;
-                    }
-                } else {
-                    if (q->constraints[j]) {
-                        fs_binding *old = q->bb[j];
-                        q->bb[j] = fs_binding_apply_filters(q, j, q->bb[j], q->constraints[j]);
-                        fs_binding_free(old);
-                        raptor_free_sequence(q->constraints[j]);
-                        q->constraints[j] = NULL;
-                    }
-                    fs_binding_union(q, q->bb[first_in_union], q->bb[j]);
-                    fs_binding_free(q->bb[j]);
-                    q->bb[j] = NULL;
-#ifdef DEBUG_MERGE
-                    printf("B%d = B%d UNION B%d\n", first_in_union, first_in_union, j);
-                    fs_binding_print(q->bb[first_in_union], stdout);
-                    printf("\n");
-#endif
-                }
+                pri_for_union[i] = j;
+                break;
             }
         }
     }
@@ -874,6 +866,9 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
     /* run through the blocks in the correct order to do the joins */
     for (int i=q->block; i>=0; i--) {
         int start = i > 1 ? i : 1;
+        /* N.B. this loop has to increment to ensure we bind OPTIONALs in the
+         * correct relative order, otherwise OPTIONAL blocks which share variables
+         * give the wrong result */
         for (int j=start; j<=q->block; j++) {
             if (!q->bb[j]) {
 #ifdef DEBUG_MERGE
@@ -900,21 +895,45 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
                         q->bb[j] = NULL;
                     }
                 } else if (q->join_type[j] == FS_UNION) {
-                    /* It's a UNION, inner join */
 #ifdef DEBUG_MERGE
-                    printf("block join B%d [X] B%d\n", i, j);
+                    printf("block B%d is in UNION group %d\n", j, q->union_group[j]);
 #endif
-                    if (q->bb[i]) {
-                        fs_binding *nb = fs_binding_join(q, q->bb[i], q->bb[j], FS_INNER);
-                        fs_binding_free(q->bb[i]);
-                        q->bb[i] = nb;
-                        if (i == 0) q->bt = q->bb[i];
+                    /* It's a UNION */
+                    /* apply constriants now, it's too trick to delay execution */
+                    if (q->constraints[j]) {
+                        fs_binding *old = q->bb[j];
+                        q->bb[j] = fs_binding_apply_filters(q, j, q->bb[j], q->constraints[j]);
+                        fs_binding_free(old);
+                        raptor_free_sequence(q->constraints[j]);
+                        q->constraints[j] = NULL;
+                    }
+                    int un = q->union_group[j];
+                    /* if this is the primary block for this union group */
+                    if (pri_for_union[un] == j) {
+#ifdef DEBUG_MERGE
+                        printf("block join B%d [X] B%d\n", i, j);
+#endif
+                        if (q->bb[i]) {
+                            fs_binding *nb = fs_binding_join(q, q->bb[i], q->bb[j], FS_INNER);
+                            fs_binding_free(q->bb[i]);
+                            q->bb[i] = nb;
+                            if (i == 0) q->bt = q->bb[i];
+                            fs_binding_free(q->bb[j]);
+                            q->bb[j] = NULL;
+                        } else {
+                            /* Bi is empty, just replace it */
+                            q->bb[i] = q->bb[j];
+                            q->bb[j] = NULL;
+                        }
+                    } else {
+                        fs_binding_union(q, q->bb[pri_for_union[un]], q->bb[j]);
                         fs_binding_free(q->bb[j]);
                         q->bb[j] = NULL;
-                    } else {
-                        /* Bi is empty, just replace it */
-                        q->bb[i] = q->bb[j];
-                        q->bb[j] = NULL;
+#ifdef DEBUG_MERGE
+                        printf("B%d = B%d UNION B%d\n", pri_for_union[un], pri_for_union[un], j);
+                        fs_binding_print(q->bb[pri_for_union[un]], stdout);
+                        printf("\n");
+#endif
                     }
                 } else if (q->join_type[j] == FS_LEFT) {
                     /* It's an OPTIONAL, left join */
@@ -936,6 +955,17 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
                     if (i == 0) q->bt = q->bb[i];
                     fs_binding_free(q->bb[j]);
                     q->bb[j] = NULL;
+                } else if (q->join_type[j] == FS_MINUS) {
+                    /* It's a MINUS, subtract RHS from LHS */
+#ifdef DEBUG_MERGE
+                    printf("block join B%d MINUS B%d\n", i, j);
+#endif
+                    fs_binding *nb = fs_binding_minus(q, q->bb[i], q->bb[j]);
+                    fs_binding_free(q->bb[i]);
+                    q->bb[i] = nb;
+                    if (i == 0) q->bt = q->bb[i];
+                    fs_binding_free(q->bb[j]);
+                    q->bb[j] = NULL;
                 } else {
                     fs_error(LOG_ERR, "unknown join type joining B%d and B%d", i, j);
                 }
@@ -951,7 +981,11 @@ int fs_query_process_pattern(fs_query *q, rasqal_graph_pattern *pattern, raptor_
 void fs_query_free(fs_query *q)
 {
     if (q) {
-        if (q->rq) rasqal_free_query(q->rq);
+        if (q->rq) {
+            g_static_mutex_lock(&rasqal_mutex);
+            rasqal_free_query(q->rq);
+            g_static_mutex_unlock(&rasqal_mutex);
+        }
 	fs_binding_free(q->bb[0]);
 	if (q->resrow) free(q->resrow);
 	if (q->ordering) free(q->ordering);
@@ -975,6 +1009,7 @@ void fs_query_free(fs_query *q)
             g_free(it->data);
         }
 	g_slist_free(q->free_list);
+        fs_query_free_row_freeable(q);
 
         if (q->default_graphs) fs_rid_vector_free(q->default_graphs);
 
@@ -1200,17 +1235,24 @@ static void graph_pattern_walk(fsp_link *link, rasqal_graph_pattern *pattern,
     int union_sub = 0;
 
     int op = rasqal_graph_pattern_get_operator(pattern);
+    int handled = 0;
 
-    if (op == RASQAL_GRAPH_PATTERN_OPERATOR_OPTIONAL) {
+    switch (op) {
+    case RASQAL_GRAPH_PATTERN_OPERATOR_OPTIONAL:
+        handled = 1;
 	(q->block)++;
         q->parent_block[q->block] = parent;
         q->join_type[q->block] = FS_LEFT;
-    } else if (op == RASQAL_GRAPH_PATTERN_OPERATOR_UNION) {
+        break;
+    case RASQAL_GRAPH_PATTERN_OPERATOR_UNION:
+        handled = 1;
         (q->unions)++;
         union_sub = q->unions;
-    } else if (op == RASQAL_GRAPH_PATTERN_OPERATOR_BASIC ||
-               op == RASQAL_GRAPH_PATTERN_OPERATOR_GRAPH ||
-               op == RASQAL_GRAPH_PATTERN_OPERATOR_GROUP) {
+        break;
+    case RASQAL_GRAPH_PATTERN_OPERATOR_BASIC:
+    case RASQAL_GRAPH_PATTERN_OPERATOR_GRAPH:
+    case RASQAL_GRAPH_PATTERN_OPERATOR_GROUP:
+        handled = 1;
         if (op == RASQAL_GRAPH_PATTERN_OPERATOR_GRAPH) {
             model = rasqal_graph_pattern_get_origin(pattern);
             if (!model) {
@@ -1220,7 +1262,9 @@ static void graph_pattern_walk(fsp_link *link, rasqal_graph_pattern *pattern,
         (q->block)++;
         q->parent_block[q->block] = parent;
         q->join_type[q->block] = FS_INNER;
-    } else if (op == RASQAL_GRAPH_PATTERN_OPERATOR_FILTER) {
+        break;
+    case RASQAL_GRAPH_PATTERN_OPERATOR_FILTER:
+        handled = 1;
         rasqal_expression *e =
             rasqal_graph_pattern_get_filter_expression(pattern);
         if (e) {
@@ -1243,7 +1287,15 @@ static void graph_pattern_walk(fsp_link *link, rasqal_graph_pattern *pattern,
 #endif
             raptor_sequence_push(q->constraints[parent], e);
         }
-    } else {
+        break;
+    case RASQAL_GRAPH_PATTERN_OPERATOR_MINUS:
+        handled = 1;
+	(q->block)++;
+        q->parent_block[q->block] = parent;
+        q->join_type[q->block] = FS_MINUS;
+        break;
+    }
+    if (!handled) {
 	fs_error(LOG_ERR, "Unknown GP operator %d not supported", op);
     }
 
@@ -1878,8 +1930,10 @@ static fs_rid const_literal_to_rid(fs_query *q, rasqal_literal *l, fs_rid *attr)
 	    return fs_hash_literal(l->value.integer ?
 			"true" : "false", *attr);
 	case RASQAL_LITERAL_INTEGER:
+        *attr = fs_c.xsd_integer;
+	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_INTEGER_SUBTYPE:
-            *attr = fs_c.xsd_integer;
+            *attr = fs_c.xsd_int;
 	    return fs_hash_literal((char *)l->string, *attr);
 	case RASQAL_LITERAL_DOUBLE:
             *attr = fs_c.xsd_double;
@@ -1987,6 +2041,11 @@ int fs_bind_slot(fs_query *q, int block, fs_binding *b,
 			"true" : "false", fs_c.xsd_boolean));
 	    break;
 	case RASQAL_LITERAL_INTEGER_SUBTYPE:
+	    if (!lit_allowed) {
+		return 1;
+	    }
+	    fs_rid_vector_append(v, fs_hash_literal((char *)l->string, fs_c.xsd_int));
+	    break;
 	case RASQAL_LITERAL_INTEGER:
 	    if (!lit_allowed) {
 		return 1;
