@@ -348,15 +348,17 @@ static void http_query_worker(gpointer data, gpointer user_data)
        http_send(ctxt, "\n");
        w = w->next;
     }
-    http_close(ctxt);
     fs_query_free(ctxt->qr);
     ctxt->qr = NULL;
-    free(ctxt->query_string);
-    ctxt->query_string = NULL;
+    if (ctxt->query_string) {
+      free(ctxt->query_string);
+      ctxt->query_string = NULL;
+    }
     if (ctxt->output) {
       g_free(ctxt->output);
       ctxt->output = NULL;
     }
+    http_close(ctxt);
 
     return;
   }
@@ -393,6 +395,8 @@ static void http_query_worker(gpointer data, gpointer user_data)
       type = "json";
     } else if (accept && strstr(accept, "text/tab-separated-values")) {
       type = "text";
+    } else if (accept && strstr(accept, "text/csv")) {
+      type = "csv";
     } else if (accept && strstr(accept, "text/plain")) {
       type = "text";
       fprintf(fp, "Content-Type: text/plain; charset=utf-8\r\n\r\n");
